package ipaudit_config;

use 5.004;
use strict;
# use warnings;
use vars qw($VERSION @ISA @EXPORT);

require Exporter;

# our @ISA = qw(Exporter);
# our @EXPORT = qw(ipa_getconf);
# our $VERSION = '0.03';

@ISA = qw(Exporter);
@EXPORT = qw(ipa_getconf);
$VERSION = '0.03';

sub ipa_getconf {
  my $Value;
  my $Name;
  my %Conf;
  my $IP_HOME=`grep ipaudit /etc/passwd | awk -F: '{print \$6}'`;
  chomp $IP_HOME;
  if ($IP_HOME eq "") {$IP_HOME = "/home/ipaudit";}
  open(IPAUDITCONF, "< $IP_HOME/ipaudit-web.conf") or 
  	open(IPAUDITCONF, "< $IP_HOME/ipaudit.conf") or 
	die "Can't Open $IP_HOME/ipaudit-web.conf or $IP_HOME/ipaudit.conf $!\n";
  while(<IPAUDITCONF>)
  {
    next if $_ =~ /^#/ || $_ =~ /^\s*\n$/; # skip comments, blanks and newlines.
    ($Name,$Value) = split(/=\s*/);        # this outta grab variables with spaces in 'em, ie: GNUPLOT
    $Name =~ s/\s//g;			   # tidy up var names with spaces infront or behind.
    $Value =~ s/"|'//g;                    # more tidy with " and '
    chomp($Conf{uc $Name} = $Value);
  }
  close(IPAUDITCONF);
  return(%Conf);
}

1;
__END__

=head1 NAME

ipaudit_config - Reads configuration values from ipaudit-web.conf

=head1 SYNOPSIS

#!/usr/bin/perl

BEGIN {
  unshift (@INC,"/path/to/installed/ipaudit_config.pm");
}

use ipaudit_config;

%conf = ipa_getconf();

# note CAPITAL {VARS} below

$zgrep = $conf{'ZGREP'}; # or whatever

=head1 DESCRIPTION

Reads in all conf values for ipaudit-web perl scripts.

=head2 EXPORT

Exports ipa_getconf() 

=head1 CHANGELOG

v0.01, 01/19/02:
  - Initial module

v0.02, 01/31/02:
  - Rewrote parsing engine to appropriately handle
    variables that include spaces.
  - Variable normalizing.
  - Updated SYNOPSIS.

v0.03, 02/04/02:
  - Overlooked the "use 5.006" - Such a new version isn't
    really needed, so it was changed to 5.004.
  - Moved "our" stuff into "use vars." Not really needed,
    and enables this to be used on older versions of perl.

=head1 AUTHOR

jh <jh@dok.org> Use and redistribution allowed at user's own
risk.

=cut
