/*
------------------------------------------------------------------------
Includes
------------------------------------------------------------------------
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*
------------------------------------------------------------------------
Defines
------------------------------------------------------------------------
*/
#define MAX_COL 256
#define NSTR    256
#define IPLEN    16

/*
------------------------------------------------------------------------
Debug macros
------------------------------------------------------------------------
*/
#define WRITEMSG \
  printf ("File %s line %d: ", __FILE__, __LINE__); \
  fflush(stdout);

#define WRITETXT(txt) \
   printf ("File %s line %d: ** %s **\n", __FILE__, __LINE__, (txt));

#define WRITEVAR(VAL,FMT) \
		   printf ("File %s line %d: ", __FILE__, __LINE__); \
		   printf ("%s=",#VAL); printf (#FMT, VAL); printf ("\n"); \
		   fflush(stdout);

/*
------------------------------------------------------------------------
Global variables
------------------------------------------------------------------------
*/
extern int errno;

/*
------------------------------------------------------------------------
Modules variables
------------------------------------------------------------------------
*/
int debug_m = 0;

/*
------------------------------------------------------------------------
Prototypes
------------------------------------------------------------------------
*/
int   get_fields     (char *s, char **fptr, int);
int   is_whitespace  (int c);
void  parse_ip_range (char *arg_in, char (**iplist)[IPLEN], int *niplist);
int   in_iprange     (char *, char (*iplist)[IPLEN], int niplist);
int   str2ip         (char *ipstr);
void  ip2str         (int ip, char[IPLEN]);
void  ip2ipf         (char *ip, char ipf[IPLEN]);
int   str2mask       (char *ipstr);
void  Print_Usage    (void);

/*
------------------------------------------------------------------------
Main
------------------------------------------------------------------------
*/
int main (int argc, char *argv[]) {
	int niplist;
	char (*iplist)[IPLEN] = NULL;
	int nfield;
	char *fptr[MAX_COL];
	int  loc0, loc1;
	char buffer[NSTR];
	char buffer2[NSTR];
	int  nbuffer;
	FILE *fin;

	if (argc==1) {
		Print_Usage();
		return 1;
	}

	parse_ip_range (argv[1], &iplist, &niplist);

	if (argc>2) {
		if (!strcmp("-",argv[2])) {
			fin = stdin;
		} else {
			fin = fopen(argv[2],"rt");
			if (NULL==fin) {
				fprintf (stderr, "ERROR:  Cannot open input file\n");
				return 1;
			}
		}
	} else
		fin = stdin;


	/*  Read ipaudit output from standard input and re-order ip data  */
   while (NULL!=fgets(buffer,NSTR,fin)) {

		/*  Skip comments  */
		if (buffer[0]=='#') continue;

		/*  Remove trailing \n  */
		nbuffer = strlen(buffer)-1;
		buffer[nbuffer] = '\0';
		strncpy(buffer2,buffer,nbuffer+1);

		/*  Break into fields  */
		nfield = get_fields (buffer, fptr, 2);

		/*  Determine if first,second ip address is local  */
		loc0 = in_iprange (fptr[0], iplist, niplist);
		loc1 = in_iprange (fptr[1], iplist, niplist);

		/*  Address already local/remote  */
		if (loc0 && !loc1) {
			printf ("%s\n", buffer2);

		/*  Reorder remote/local  */
		} else if (!loc0 && loc1) {
			nfield = get_fields (buffer2, fptr, MAX_COL);
			/* do stuff  */
			printf ("%s %s %s %s %s %s %s %s %s %s %s %d %d\n",
					fptr[1],
					fptr[0],
					fptr[2],
					fptr[4],
					fptr[3],
					fptr[6],
					fptr[5],
					fptr[8],
					fptr[7],
					fptr[9],
					fptr[10],
					3-atoi(fptr[11]),
					3-atoi(fptr[12])
					);
		}
		/*  Local/Local or Remote/Remote are rejected  */
	}

	/*  Close file  */
	if (argc>2)
		fclose (fin);
}



/*
------------------------------------------------------------------------
Local functions
------------------------------------------------------------------------
*/
void parse_ip_range (char *arg_in, char (**iplist)[IPLEN], int *niplist) {
	char *arg_cpy = (char *) malloc (strlen(arg_in)+1);
	char *ipstr   = (char *) malloc (strlen(arg_in)+1);
	char *netstr  = (char *) malloc (strlen(arg_in)+1);
	char *range1  = NULL;
	char *range2  = NULL;
	int  nrange;
	int  mask;
	int  net;
	int  ip1, ip2;
	int  n;
	char *p;

	*iplist = NULL;

	/*  Count number of ranges (equals number of : + 1 )  */
	p = arg_in;
	n = 1;
	while (*p++) {
		if (*p==':') n++;
	}

	/*  allocate storage  */
	*iplist = (char (*)[IPLEN]) malloc (2 * n * IPLEN);
	if (*iplist==NULL) {
		*niplist = 0;
		return;
	}

	strcpy  (arg_cpy, arg_in);
	range2 = arg_cpy;

	/*  break string into separate ranges  */
	*niplist = 0;
	while (NULL!=range2) {

		/*  Break arg into (1st range):(remaining ranges)  */
		range1 = range2;
		range2 = strchr(range1, ':');
		if (NULL!=range2) *range2++ = '\0';


		/*  Look for range expressed as (lo ip)-(hi ip)  */
	 	if (2==sscanf (range1, "%[0-9.]-%[0-9.]", ipstr, netstr)) {
			ip1 = str2ip(ipstr);
			ip2 = str2ip(netstr);

		/*  break range into (ip)/(net)  */
		} else if (2==sscanf (range1, "%[0-9.]/%[0-9]", ipstr, netstr)) {

			/*  read ip address  */
			ip1 = str2ip (ipstr);

			net = atoi(netstr);
			if (net<0) net=0;
			else if (net>32) net=32;
			mask = 0xffffffff >> net;
			if (mask==-1) mask = 0;
			ip2 = ip1 | mask;

		/*  Look for single ip address  */
		} else if (sscanf (range1, "%[0-9.].%[0-9].", ipstr, netstr)) {
			ip1 = str2ip (ipstr);
			ip2 = ip1 | str2mask(ipstr);

		/*  Bad input format  */
		} else {
			fprintf (stderr, 
					"ERROR:  Cannot read network range argument (-l option).\n");
			fprintf (stderr, 
					"  Program continues with using default network range.\n");
			*niplist = 0;
			if (NULL!=*iplist) free (*iplist);
			return;
		}

		/* Store results  */
		ip2str (ip1,(*iplist)[(*niplist)++]);
		ip2str (ip2,(*iplist)[(*niplist)++]);
	} 


	free (netstr);
	free (ipstr);
	free (arg_cpy);


	/*  Correct double counting of niplist  */
	*niplist /= 2;

}


/*  Determine if ipaddresses is within one of the ranges in iplist  */
int in_iprange (char *ip, char (*iplist)[IPLEN], int niplist) {
	int i;
	char ipf[IPLEN];
	ip2ipf(ip,ipf);
	for (i=0;i<2*niplist;i+=2) 
		if (strcmp(ipf,iplist[i])>=0 && strcmp(ipf,iplist[i+1])<=0)
			return 1;
	return 0;
}

int get_fields (char *s, char **fptr, int max_field) {
	int nfield=0;

   /*  Skip leading whitespace  */
   while (*s==' ' || *s=='\t') s++;
   /*  Find tokens  */
   while (*s && nfield<max_field) {
      fptr[nfield++] = s;
      /*  Find first whitespace  */
		while (*s!=' ' && *s && *s!='\t')  s++;
		/*  Make first white space string terminator */
		if (*s)  *s++ = '\0';
		/*  Find last whitespace */
		while (*s==' ' || *s=='\t')  s++;
	}
	return nfield;
}

/*  Convert strings like "138.99.201.5" or "137.99.26" to int ipaddress  */
int str2ip (char *ipstr) {
   int ip[4];
   int i;
	int ipout=0;
   int n = sscanf (ipstr, "%d.%d.%d.%d", ip, ip+1, ip+2, ip+3);
   for (i=0;i<4;i++) {
      ipout = ipout<<8;
      if (i<n) ipout |= (ip[i] & 0xff);
   }
	return ipout;
}


int str2mask (char *ipstr) {
	int ip[4];
	int mask;
	int n = sscanf (ipstr, "%d.%d.%d.%d", ip, ip+1, ip+2, ip+3);
	mask = 0xffffffff >> (8*n);

	/* for reasons unknown 0xffffffff >> 32 -> -1, so set to 0  */
	if (mask==-1)  mask=0;

	return mask;
}


void Print_Usage (void) {
	printf ("\nmakelocal <network> [file]\n\n");
	printf ("   Converts ipaudit data file to make <network> the\n");
	printf ("   local network.  The format for the <network> strings\n");
	printf ("   is identical to that for ipaudit's -l option\n");
	printf ("   (see ipaudit man page).  Converted data is written to\n");
	printf ("   standard out.  If file is omitted or '-' is used then\n");
	printf ("   ipaudit data is read from standard in.\n\n");
}


/*  Convert int ip to zero padded string such as 001.231.021.124  */
void ip2str (int ip, char buffer[IPLEN]) {
	int p[4];
	int i;
	for (i=0;i<4;i++) {
		p[i] = ip & 0xff;
		ip >>= 8;
	}
	sprintf (buffer, "%03d.%03d.%03d.%03d", p[3], p[2], p[1], p[0]);
}

/*  Copy ip string into buffer with zero padding  */
void ip2ipf (char *ip, char ipf[IPLEN]) {
	int slen = strlen(ip);
	int olen;
	char *dst, *src;
	/*  return if ip string already padded  */
	if (slen>=IPLEN-1) {
		strncpy (ipf, ip, IPLEN);
		return;
	}
	src = ip  + slen  - 1;
	dst = ipf + IPLEN - 1;
	*(dst--) = '\0';
	/*  Work forward from end of string  */
	while (dst>=ipf) {
		olen = 3;
		/*  Transfer digits of octet  */
		while (*src!='.' && src>=ip) {
			*(dst--) = *(src--);
			olen--;
		}
		/*  Pad octet with zeroes  */
		while (olen--)
			*(dst--) = '0';
		/*  Add '.'  */
		if (dst>ipf)
			*(dst--) = '.';
		src--;
	}
}
